# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-08-28 12:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LDPSource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('container', models.URLField()),
                ('federation', models.CharField(max_length=255)),
            ],
            options={
                'ordering': ('federation',),
                'permissions': (('view_source', 'acl:Read'), ('control_source', 'acl:Control')),
            },
        ),
    ]
